/*
** Module   : AdaptiveQuery.java
** Abstract : Query which adapts its retrieval type from preselect to dynamic
**
** Copyright (c) 2004-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060721   @28133 Created initial version. Query which adapts
**                           its retrieval type from preselect to dynamic,
**                           based upon DMO state change notifications.
** 002 ECF 20060725   @28186 Integrated ProgressiveResults. Implementation
**                           provides better performance for use cases
**                           which require access to a small number of
**                           result records while retaining most of the
**                           preselect performance advantage for larger
**                           queries.
** 003 ECF 20060730   @28262 Fixed order by clause defect. Order by clause
**                           was being omitted.
** 004 ECF 20060801   @28277 Fixed NPE in order by processing. Added
**                           safety code to assembleOrderByClause().
** 005 ECF 20060802   @28347 Enabled transition from dynamic to preselect
**                           mode. Once invalidated, an AdaptiveQuery will
**                           now transition back to preselect mode as soon
**                           as possible.
** 006 ECF 20060803   @28387 Integrated ResultsAdapter and made this class
**                           more easily extended. Refactored much of the
**                           behavior into polymorphic methods.
** 007 ECF 20060808   @28573 Refined ResultsAdapter to better handle
**                           FilteredResults. Integrated ResultsProvider
**                           interface with new inner class ResultsSource.
** 008 ECF 20060830   @28992 Fixed join behavior. For a single table query
**                           and for the first table of a multi-table
**                           query, a natural join's inverse DMO is
**                           assumed to be provided as a query parameter,
**                           and a DynamicJoin is created. For all other
**                           natural joins, the ServerJoin implementation
**                           is used.
** 009 GES 20060830   @29018 Fixed null ptr exception in tracing code.
** 010 ECF 20060831   @29082 Added support for sorting by an indexed
**                           property. Specifically, support sorting by
**                           a property of a DMO's associated list of
**                           composite properties.
** 011 ECF 20060901   @29154 Fixed NPE. Cannot safely use invalidBuffer to
**                           extract DMO proxy in createSimpleQuery().
** 012 ECF 20060911   @29459 Fixed defect in transition from dynamic to
**                           preselect mode. A state change event during
**                           this transition must cancel the transition.
** 013 ECF 20061006   @30195 Implemented scrolling mode. When set to
**                           scrolling mode, non-linear scrolling is
**                           possible in preselect mode. Previously, only
**                           linear scrolling from first to last was
**                           possible. This is necessitated by the fact
**                           that AdaptiveQuery may back a browse widget.
** 014 ECF 20061019   @30524 Additional fixes to support scrolling mode.
**                           These were made primarily to support the use
**                           of this query with a Browse widget, and
**                           possibly as a component within a scrolling
**                           CompoundQuery.
** 015 ECF 20061027   @30748 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 016 ECF 20061119   @31323 Refined invalidation strategy. Do not switch
**                           from preselect mode on session closing events
**                           if a reposition listener is registered to the
**                           query (indicates a browse association). Upon
**                           record change events, only invalidate if
**                           changed property is one which would affect
**                           content or sorting of result set. Also
**                           improved dynamic mode reposition fetching.
** 017 ECF 20061121   @31343 Added safety code to prevent NPE. During
**                           setupInvalidationProperties(), query
**                           component may return null map of restriction
**                           property names if where clause was null.
** 018 ECF 20061129   @31477 Additional change related to #016 (@31323).
**                           If associated with a browse on session close,
**                           discard current high level results object.
** 019 ECF 20061203   @31573 Further changes to sessionClosing() and
**                           invalidate(RecordBuffer) methods. We now
**                           invalidate results in the event of an error
**                           on session close. We also use the explicit
**                           browse check (isBrowsed()) to detect whether
**                           the query is associated with a browse when
**                           invalidating results. Also, this test was
**                           moved from sessionClosing() to invalidate(),
**                           to ensure that a browse association was being
**                           considered in all invalidation scenarios, not
**                           just session close.
** 020 ECF 20061207   @31617 Support insert RecordChangeEvent. Modified
**                           stateChanged() method.
** 021 ECF 20070223   @32226 Do not switch to dynamic mode if the record
**                           currently being visited is deleted. This
**                           supports a common idiom of deleting records
**                           in a loop without thrashing back and forth
**                           between modes.
** 022 ECF 20070227   @32242 Fixed #021 (@32226). The implementation of
**                           this optimization was incorrect.
** 023 ECF 20070330   @32242 Fixed NPE in stateChanged() method.
** 024 ECF 20070412   @32974 Fixed createSimpleQuery(). Always use lock
**                           type of NONE for the dynamic query. This will
**                           be overridden by PreselectQuery.fetch(), if
**                           a stricter lock is required.
** 025 ECF 20070416   @33017 Integrated user interrupt handling.
** 026 ECF 20070509   @33452 Modified sessionClosing() behavior to behave
**                           as superclass if scrolling. Temporarily
**                           rolled back #021 (@32226), which causes
**                           ProgressiveResults to lose records. Need a
**                           better solution for this problem which
**                           preserves the optimization in #021.
** 027 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 028 ECF 20070706   @34400 Minor optimization. Replaced StringBuffer
**                           with StringBuilder.
** 029 ECF 20070709   @34421 Modified sessionClosing(). Always preserve
**                           preselected results rather than invalidating
**                           the query. This significantly improves
**                           performance in certain use patterns where the
**                           session is closed regularly before the query
**                           has iterated all of its results (e.g., a
**                           commit which occurs within a loop at a lower
**                           nested scope).
** 030 ECF 20070711   @34445 Further improved sessionClosing(). Event
**                           handling on session close is now delegated to
**                           the results adapter.
** 031 ECF 20070720   @34905 Added toString() method for debugging.
** 032 ECF 20070827   @34998 Handle temp table multiplex ID when used in
**                           query's sort criteria. Certain logic had an
**                           improper dependency on all SortCriterion
**                           instances having a getter method, which is
**                           not the case for multiplex IDs.
** 033 ECF 20071022   @35557 Prevent thrashing between modes. Added crude
**                           safety mechanism which disables transitioning
**                           from dynamic to preselect mode more than 3
**                           times.
** 034 ECF 20071130   @36118 Integrated generics.
** 035 ECF 20071206   @36270 Removed sort field from AdaptiveComponent.
**                           This field was not being used. Constructor
**                           now sets SortIndex in superclass.
** 036 ECF 20080122   @36888 Fixed handling of state change events in
**                           dynamic mode. It is necessary to reset the
**                           dynamic query's cursor to avoid retrieving
**                           stale data.
** 037 SVL 20071230   @36222 When ProgressiveResults reports that changes
**                           in results detected while getting them, the
**                           mode is switched from preselect to dynamic.
** 038 SVL 20080418   @38040 Query registers as ChangeBroker listener in
**                           the constructor.
** 039 ECF 20080502   @28133 Removed isRepositionOnOpen(). Method is
**                           obsolete.
** 040 SVL 20080526   @38312 isPrelockUsed() added.
** 041 SVL 20080604   @38535 currentRowImpl() is used instead of
**                           currentRow().
** 042 CA  20080617   @38867 If there are components with no inverse
**                           record to join on, then the query will not
**                           be executed and no results will be returned.
** 043 ECF 20080701   @39033 Optimized stateChanged(). Avoid invalidating
**                           preselected results in certain delete and
**                           update cases when we can detect invalidation
**                           is unnecessary.
** 044 ECF 20080708   @39081 Refined #043 (@39033) optimization. A delete
**                           of the first record found by the query will
**                           trigger an immediate query/caching of all
**                           remaining results. Tightened conditions where
**                           double-checking of invalidation is done.
** 045 ECF 20080709   @39085 Fixed sessionClosing() and stateChanged(). We
**                           now prevent using a dead connection after the
**                           session closes and avoid caching results on a
**                           delete event in dynamic mode, respectively.
** 046 ECF 20080710   @39091 Modified sessionClosing(). cacheResults()
**                           no longer throws IllegalStateException.
** 047 ECF 20080723   @39174 Added retry handling. Implement Finalizable
**                           and prepare for retry in retry() method by
**                           setting state as if a reposition to the
**                           current record has occurred.
** 048 CA  20080729   @39219 Removed retry handling. Retry was added back 
**                           to PreselectQuery.
** 049 SVL 20080730   @39223 Support API change in IndexHelper.
** 050 SVL 20080815   @39400 Into provideNewResults() before we skip the
**                           first result got after switching to the
**                           preselect mode, we check whether it equals
**                           the last result retrieved in dynamic mode.
** 051 ECF 20080813   @39553 Support invalidation of results based on DMO
**                           change events in other sessions. Currently,
**                           this support is limited to changes which
**                           affect an index on the backing table.
** 052 CA  20080815   @39449 Support API change in DBUtils.
** 053 ECF 20081014   @40088 Added support for inlining query substitution
**                           parameters. Integrated HQLPreprocessor API
**                           changes;  modified AdaptiveComponent to use
**                           appropriate HQL, based on preselect vs.
**                           dynamic mode.
** 054 ECF 20081030   @40305 Fixed regression introduced by #053 (@40088).
**                           Refactored AdaptiveComponent inner class and
**                           code in top-level class which uses it.
** 055 ECF 20081105   @40309 Overrode parent's open() method. Registers
**                           with the ChangeBroker at the appropriate
**                           scope.
** 056 CA  20081210   @40836 Do not cleanup if the query was explicitly
**                           closed.
** 057 CA  20081211   @40858 Fixed regression introduced by H056: on
**                           sessionClosing event, if the query is closed,
**                           return true.
** 058 ECF 20081216   @40916 Removed flushBuffers() invocation. This work
**                           is now done when a component is added to the
**                           query. The previous invocation, during record
**                           retrieval, was too late.
** 059 SVL 20090118   @41169 getOriginalWhere() is used together with
**                           getRawArgs() during dynamic query creation.
** 060 ECF 20090202   @41247 Rolled back #040 (@38312). Prelock mechanism
**                           has been disabled.
** 061 ECF 20090220   @41350 Override parent's getRow() method. Needed in
**                           cases where row data is required during a
**                           transition from preselect to dynamic mode,
**                           before a new Results object is available.
** 062 ECF 20090630   @42986 Fixed reposition support in conjunction with
**                           changes to the base class.
** 063 ECF 20090701   @43008 Fixed NPE in setupInvalidationProperties(). It is
**                           possible for a component to have null sort
**                           criteria, if the query's only sort is by primary
**                           key.
** 064 ECF 20090702   @43033 Added isNativelyScrolling(). Overrides parent's
**                           implementation to report false.
** 065 ECF 20090703   @43053 Reimplemented load() method. This method now sets
**                           the backing buffer to unknown mode and throws
**                           MissingRecordException if a target record cannot
**                           be loaded.
** 066 ECF 20090707   @43102 Fixed regression in #062. When storing results in
**                           the ResultsAdapter in createResults(), set the
**                           betweenRows flag to false to ensure the next call
**                           to next() must advance the query.
** 067 ECF 20090716   @43212 Added late initialization of backing record
**                           buffer.
** 068 ECF 20090724   @43414 Added implicit transaction support. Added
**                           push/pop implicit transaction calls to
**                           DynamicResults.
** 069 CA  20090731   @43464 Register the query for off-end listener 
**                           registration after block initialization (done
**                           only for block types FOR and FOR EACH).
** 070 ECF 20090810   @43585 Changed to accommodate SessionListener API
**                           change. sessionClosing() method was replaced with
**                           sessionEvent().
** 071 ECF 20090811   @43597 Fixed resource cleanup. Made changes to execute()
**                           to make sure a resource cleaner is registered in
**                           dynamic mode. Added call to cleanup adapter when
**                           switching modes.
** 072 SVL 20091030   @44303 Added isNativelyPreselect() function.
** 073 SVL 20120828          Make dynamic results non-ID-only if there is a
**                           client-side where clause.
** 074 ECF 20131028          Import change.
** 075 SVL 20140210          Use HQLExpression instead of HQL string.
** 076 SVL 20140604          Added support for INDEX-INFORMATION.
** 077 SVL 20140812          Fixed PREV mistakenly treated as LAST.
** 078 SVL 20140909          Added cursor for proper support of scrolling queries.
**                           Implemented deleteResultListEntry.
** 079 SVL 20141209          Fixed case of full and empty cursor set.
** 080 SVL 20150212          Use normal invalidation for browsed queries. Removed duplicate
**                           functionality in repositionByID (it is handled by Cursor).
** 081 SVL 20150331          Added deleteResultListEntry(boolean).
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import org.apache.commons.logging.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.event.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;

/**
 * An adaptive, iterating query which first attempts to retrieve its records
 * in preselect mode, but changes to dynamic record retrieval as necessary.
 * <code>AdaptiveQuery</code> balances the high performance of a preselect
 * query implementation with the flexibility of a dynamic query
 * implementation.
 * <p>
 * The design point behind this implementation is that in many if not most
 * cases, a preselect query plan is the highest performing strategy.  It
 * leverages the strength of the database server to do the heavy lifting of
 * set-oriented data record retrieval.  In most use cases, this is the most
 * appropriate retrieval strategy.  However, to emulate Progress' semantic,
 * it is necessary to reflect table updates in a query which has already
 * begun iterating through its result set.  This dynamic approach is much
 * slower in the P2J architecture, since it requires at minimum one, and
 * sometimes many more round trips to the server for each record retrieved.
 * Therefore it is desirable to avoid dynamic retrieval mode whenever
 * possible.
 * <p>
 * The one performance weakness of the preselect approach is the case where
 * client code does not need the full result set acquired by the query.  To
 * deal with this, <code>AdaptiveQuery</code> uses an instance of {@link
 * ProgressiveResults} to fetch progressively larger brackets of result set
 * data while in preselect mode.  Thus, early results are returned quickly,
 * and if more results are needed, they are fetched in increasingly larger
 * batches.  This amortizes the access cost most effectively for both ends of
 * the spectrum:  use cases which require only one or a handful of results
 * never incur the cost of larger fetches and use cases which require many
 * results acquire them with far fewer calls than in dynamic mode.
 * <p>
 * <code>AdaptiveQuery</code> takes an optimistic approach by operating in
 * preselect mode immediately and for as long as possible.  It registers as a
 * {@link ChangeBroker} listener.  If during its preselect mode operation it
 * receives notification of a DMO state change which would invalidate its
 * current result set, it automatically switches over to dynamic retrieval
 * mode.  It will remain in dynamic mode for as long as necessary, but will
 * switch back into preselect mode at the earliest opportunity.
 * <p>
 * Once switching from preselect to dynamic mode, the query will operate in
 * this mode until the underlying <code>DynamicResults</code> object detects
 * that it is possible to switch back to preselect mode.  This occurs when
 * the <code>DynamicQuery</code> underlying the <code>DynamicResults</code>
 * object crosses a <i>sort band</i> boundary.  A <i>sort band</i> is the
 * grouping defined by the coarsest sort criterion of the query.  A boundary
 * is crossed when the DMO property defining the sort band changes from one
 * retrieved record to the next.  The value of this property in the last
 * record before the boundary is crossed is termed the <i>break value</i>.
 * When the break value for the underlying dynamic query is non-null, this is
 * an indication that the sort band boundary has been crossed.  At this point,
 * it is possible to construct a new preselect mode query which uses all the
 * original query criteria, plus the break value, to ensure the query results
 * pick up where the dynamically retrieved result set ended.
 * <p>
 * <b>WARNING:</b>  currently, the multi-table implementation is not
 * functional.
 * <p>
 * TODO:  support multiple table queries.  The support is partially in place
 * in the <code>DynamicQuery</code> inner class;  however, the instance of
 * <code>CompoundQuery</code> which is created to drive the continuation of
 * the query after invalidation currently does not have its state set
 * correctly.
 * <p>
 * TODO:  Make invalidation check more robust;  currently it is too
 * conservative.
 * <p>
 * TODO:  Make revalidation failsafe more sophisticated;  currently it locks
 * into dynamic mode after 3 revalidations.
 * <p>
 * TODO:  Make global invalidation setting configurable from the directory.
 */
public class AdaptiveQuery
extends PreselectQuery
implements RecordChangeListener,
           VirtualResultsListener
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(AdaptiveQuery.class);
   
   /** Force invalidation checks across all sessions */
   private static final boolean forceGlobalInvalidation = false;
   
   /** Results adapter which can connect to different low-level delegates */
   private final ResultsAdapter adapter;
   
   /** Implement invalidation checks across all sessions */
   private boolean globalInvalidation = forceGlobalInvalidation;
   
   /** High level results object for parent query */
   private Results highLevelResults = null;
   
   /** Query component which triggered invalidation */
   private AdaptiveComponent invalidComponent = null;
   
   /** Properties whose changes trigger invalidation, keyed by DMO entity */
   private Map<String, Set<String>> invalidationProperties = null;
   
   /** Property value which indicates a new sort band in dynamic mode */
   private Object breakValue = null;
   
   /** Number of times this query has transitioned back to preselect mode */
   private int revalCount = 0;
   
   /** Are results cached due to a previously committed transaction? */
   private boolean resultsCached = false;
   
   /** Preselected results should all be cached on next revalidation */
   private boolean cacheOnReval = false;
   
   /** Reflects whether an error was detected while getting results */
   protected boolean errorCondition = false;

   /** Represents cache of the dynamic scrolling query. */
   protected Cursor cursor = null;

   /** Determines if we have fetched a result row. Only for non-scrolling queries. */
   protected boolean foundFirst = false;

   /**
    * Constructor which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    * <p>
    * Each component will provide its own sort criteria;  they are combined
    * when the query is first executed in preselect mode.
    */
   public AdaptiveQuery()
   {
      super(null);
      
      adapter = new ResultsAdapter(new ResultsSource());
      
      ChangeBroker.get().addListener(this);

      TransactionManager.registerOffEndQuery(this);
   }
   
   /**
    * Constructor which is typically used for a multi-table query. Details
    * specific to each table component of the query are added via one of the
    * <code>addComponent</code> method variants.
    * <p>
    * This variant is used by code which is converted from an OPEN QUERY
    * statement, which permits advancing off the end of a result set without
    * raising a <code>QueryOffEndException</code>.
    *
    * @param   lenientOffEnd
    *          <code>true</code> if advancing past either end of the result
    *          set should NOT raise a <code>QueryOffEndException</code>;
    *          <code>false</code> if doing so should raise the exception.
    */
   public AdaptiveQuery(boolean lenientOffEnd)
   {
      super(lenientOffEnd, null);
      
      adapter = new ResultsAdapter(new ResultsSource());
      
      ChangeBroker.get().addListener(this);

      TransactionManager.registerOffEndQuery(this);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort)
   {
      this(dmo, where, whereExpr, sort, (DataModelObject) null, null, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo)
   {
      this(dmo, where, whereExpr, sort, indexInfo, null, null, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        DataModelObject inverse)
   {
      this(dmo, where, whereExpr, sort, inverse, null, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.  A share lock
    * will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo,
                        DataModelObject inverse)
   {
      this(dmo, where, whereExpr, sort, indexInfo, inverse, null, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        Object[] args)
   {
      this(dmo, where, whereExpr, sort, (DataModelObject) null, args, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo,
                        Object[] args)
   {
      this(dmo, where, whereExpr, sort, indexInfo, null, args, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        DataModelObject inverse,
                        Object[] args)
   {
      this(dmo, where, whereExpr, sort, inverse, args, LockType.SHARE);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    * A share lock will be applied to retrieved records.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo,
                        DataModelObject inverse,
                        Object[] args)
   {
      this(dmo, where, whereExpr, sort, indexInfo, inverse, args, LockType.SHARE);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        LockType lockType)
   {
      this(dmo, where, whereExpr, sort, (DataModelObject) null, null, lockType);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo,
                        LockType lockType)
   {
      this(dmo, where, whereExpr, sort, indexInfo, null, null, lockType);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        DataModelObject inverse,
                        LockType lockType)
   {
      this(dmo, where, whereExpr, sort, inverse, null, lockType);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    * The where clause must expect no substitution parameters.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo,
                        DataModelObject inverse,
                        LockType lockType)
   {
      this(dmo, where, whereExpr, sort, indexInfo, inverse, null, lockType);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        Object[] args,
                        LockType lockType)
   {
      this(dmo, where, whereExpr, sort, (DataModelObject) null, args, lockType);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo,
                        Object[] args,
                        LockType lockType)
   {
      this(dmo, where, whereExpr, sort, indexInfo, null, args, lockType);
   }

   /**
    * Convenience constructor designed for a single-table adaptive query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        DataModelObject inverse,
                        Object[] args,
                        LockType lockType)
   {
      this(dmo, where, whereExpr, sort, null, inverse, args, lockType);
   }
   
   /**
    * Convenience constructor designed for a single-table adaptive query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   whereExpr
    *          Client-side where clause expression.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this query.
    */
   public AdaptiveQuery(DataModelObject dmo,
                        String where,
                        WhereExpression whereExpr,
                        String sort,
                        String indexInfo,
                        DataModelObject inverse,
                        Object[] args,
                        LockType lockType)
   {
      super(null);
      
      addComponent(dmo,
                   where,
                   sort,
                   indexInfo,
                   inverse,
                   args,
                   lockType,
                   NEXT,
                   false);
      
      adapter = new ResultsAdapter(new ResultsSource());
      
      if (whereExpr != null)
      {
         addWhereExpression(whereExpr);
      }
      
      ChangeBroker.get().addListener(this);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    */
   public void addComponent(DataModelObject dmo, String where, String sort)
   {
      addComponent(dmo,
                   where,
                   sort,
                   null,
                   null,
                   LockType.SHARE,
                   NEXT,
                   false);
   }

   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            DataModelObject inverse)
   {
      addComponent(dmo,
                   where,
                   sort,
                   inverse,
                   null,
                   LockType.SHARE,
                   NEXT,
                   false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            Object[] args)
   {
      addComponent(dmo,
                   where,
                   sort,
                   null,
                   args,
                   LockType.SHARE,
                   NEXT,
                   false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>share lock
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            DataModelObject inverse,
                            Object[] args)
   {
      addComponent(dmo,
                   where,
                   sort,
                   inverse,
                   args,
                   LockType.SHARE,
                   NEXT,
                   false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            LockType lockType)
   {
      addComponent(dmo,
                   where,
                   sort,
                   null,
                   null,
                   lockType,
                   NEXT,
                   false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no substitution parameters
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            DataModelObject inverse,
                            LockType lockType)
   {
      addComponent(dmo,
                   where,
                   sort,
                   inverse,
                   null,
                   lockType,
                   NEXT,
                   false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo,
                   where,
                   sort,
                   null,
                   args,
                   lockType,
                   NEXT,
                   false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>inner join
    *   <li><code>NEXT</code> iteration type
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType)
   {
      addComponent(dmo,
                   where,
                   sort,
                   inverse,
                   args,
                   lockType,
                   NEXT,
                   false);
   }
   
   /**
    * Add a component to the query.  The following defaults are used:
    * <ul>
    *   <li>no inverse join DMO
    * </ul>
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.
    *          Only <code>NEXT</code> is supported.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      addComponent(dmo,
                   where,
                   sort,
                   null,
                   args,
                   lockType,
                   iteration,
                   outer);
   }

   /**
    * Add a component to the query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.
    *          Only <code>NEXT</code> is supported.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      addComponent(dmo, where, sort, null, inverse, args, lockType, iteration, outer);
   }
   
   /**
    * Add a component to the query.
    *
    * @param   dmo
    *          DMO proxy which defines buffer and record type.
    * @param   where
    *          HQL where clause.  May be <code>null</code>.
    * @param   sort
    *          HQL order by clause.
    * @param   indexInfo
    *          Index information string as it is returned by INDEX-INFORMATION.
    *          May be <code>null</code> if it is not an OPEN QUERY case or if
    *          you don't need debug information about selected indexes.
    * @param   inverse
    *          DMO to which this query should join via a foreign relation.
    *          May be <code>null</code>.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          the where clause.
    * @param   lockType
    *          Lock type to apply to records retrieved by this component of
    *          the query.
    * @param   iteration
    *          Iteration/retrieval mode for this component of the query.
    *          Only <code>NEXT</code> is supported.
    * @param   outer
    *          <code>true</code> if an outer join is required.  At this time,
    *          only inner joins are supported.  Future support for outer
    *          joins is expected.
    *
    * @throws  UnsupportedOperationException
    *          if <code>outer</code> is <code>true</code> (temporary);
    *          if <code>iteration</code> is not <code>NEXT</code> (temporary);
    *          if <code>dmo</code>'s associated database is different than
    *          that of a previously added component's database (cross-database
    *          preselect joins are not possible).
    */
   public void addComponent(DataModelObject dmo,
                            String where,
                            String sort,
                            String indexInfo,
                            DataModelObject inverse,
                            Object[] args,
                            LockType lockType,
                            int iteration,
                            boolean outer)
   {
      if (getTableCount() > 0)
      {
         throw new UnsupportedOperationException(
            "Multi-table adaptive queries are not supported in this release");
      }
      
      if (iteration != NEXT)
      {
         throw new UnsupportedOperationException(
            "Adaptive query only supports the NEXT iteration type");
      }
      
      BufferReference proxy = (BufferReference) dmo;
      BufferReference inverseProxy = (BufferReference) inverse;
      AbstractJoin join = processJoin(proxy, inverseProxy);
      
      Component comp = new AdaptiveComponent(proxy,
                                             join,
                                             where,
                                             sort,
                                             indexInfo,
                                             lockType,
                                             args,
                                             iteration,
                                             outer,
                                             inverseProxy);
      addComponent(comp);
      
      // Temporary tables are local by nature, so we don't care what
      // happens in other sessions.
      if (globalInvalidation && comp.getBuffer().isTemporary())
      {
         globalInvalidation = false;
      }
   }
   
   /**
    * Open the query, first registering with the <code>ChangeBroker</code> at
    * the appropriate scope.
    */
   public void open()
   {
      int scope = TransactionManager.findNearestExternal();
      RecordBuffer[] buffers = getRecordBuffers();

      for (RecordBuffer buffer : buffers)
      {
         scope = ((scope == -1)
                  ? buffer.getScopeOpenDepth()
                  : Math.min(scope, buffer.getScopeOpenDepth()));
      }

      if (scope != -1)
      {
         registerRecordChangeListeners(scope);
      }
      
      super.open();
   }
   
   /**
    * Register all <code>RecordChangeListener</code>s associated with this
    * query (if any), with the context-local <code>ChangeBroker</code> at the
    * indicated scope.
    * 
    * @param   scope
    *          Scope at which listeners should be registered with the change
    *          broker.
    */
   public void registerRecordChangeListeners(int scope)
   {
      ChangeBroker.get().addListener(this, scope);
   }

   /**
    * Retrieve the first composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void first(LockType lockType)
   {
      if (isScrolling())
      {
         if (cursor.isFoundFirst())
         {
            Object[] data = cursor.getFirst();

            if (data == null && cursor.isFullSet())
            {
               // The set is full and empty.
               loadByValues(null, lockType);
            }
            else if (data != null)
            {
               if (!loadByValues(data, lockType))
               {
                  next(lockType);
               }
            }
            else
            {
               if (cursor.isFirstRowDeleted() && cursor.isForwardScroll())
               {
                  next(lockType);
                  return;
               }
               else
               {
                  super.first(lockType);
               }

               cursor.addResultFirst(getCurrentIds());
            }
         }
         else
         {
            super.first(lockType);
            cursor.addResultFirst(getCurrentIds());
         }
      }
      else
      {
         super.first(lockType);
      }
   }

   /**
    * Retrieve the last composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void last(LockType lockType)
   {
      if (isScrolling())
      {
         if (cursor.isFoundFirst())
         {
            Object[] data = cursor.getLast();

            if (data == null && cursor.isFullSet())
            {
               // The set is full and empty.
               loadByValues(null, lockType);
            }
            else if (data != null)
            {
               if (!loadByValues(data, lockType))
               {
                  previous(lockType);
               }
            }
            else
            {
               super.last(lockType);
               cursor.addResultLast(getCurrentIds());
            }
         }
         else
         {
            super.last(lockType);
            cursor.addResultLast(getCurrentIds());
         }
      }
      else
      {
         super.last(lockType);
      }
   }

   /**
    * Get the 1-based index of the current row in the result set.
    *
    * @return  Current row or the unknown value if the query has not yet
    *          executed.
    */
   public integer currentRowImpl()
   {
      if (isScrolling())
      {
         return cursor.currentRow();
      }
      else
      {
         return super.currentRowImpl();
      }
   }

   /**
    * Progress-compatible function that returns the 1-based index of the
    * current row in the result set.
    *
    * @return  Unknown value if the query is not open or result set is empty.
    *          1 if the query is positioned before the first record.
    *          A value 1 greater than the number of rows in the query result
    *          set if the query is positioned beyond the last record.
    *          Otherwise returns the number of rows in the query result set.
    *
    */
   public integer currentRow()
   {
      if (isScrolling())
      {
         return cursor.currentRow();
      }
      else
      {
         return super.currentRow();
      }
   }

   /**
    * Reload the current composite row of results for the query, overriding
    * the lock type to apply to each record.
    * <p>
    * If the query involves outer joins, some of the underlying buffers may
    * be empty when this method returns.
    *
    * @param   lockType
    *          Lock type which should override that of any query component.
    *
    * @throws  ErrorConditionException
    *          if the reload triggers an error.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void current(LockType lockType)
   {
      if (currentRowDeleted)
      {
         if (isScrolling() || _isOffEnd())
         {
            ErrorManager.displayError(4114, "No query record is available", false);
         }
         return;
      }

      super.current(lockType);
   }

   /**
    * Reposition the query to the specified row in the result set.  The row
    * indicates the position between two results, such that a call to
    * retrieve the next record will get the record after the new position,
    * and a call to retrieve the previous record will get the record before
    * the new position.
    *
    * @param   row
    *          1-based index of the target position.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public void reposition(int row)
   {
      if (isScrolling())
      {
         prepareReposition();

         cursor.reposition(row);
         OffEnd res = cursor.getOffEnd();
         if (!OffEnd.NONE.equals(res))
         {
            offEnd = res;
            betweenRows = false;
         }
         else
         {
            betweenRows = true;
         }

         afterReposition(true);
      }
      else
      {
         super.reposition(row);
      }
   }

   /**
    * Advance the current cursor position forward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row forward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it ahead <i>one and a half positions</i>, to just
    *       beyond the following result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it ahead <i>one position</i>, to just beyond the
    *       following result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor forward.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public void forward(int rows)
   {
      if (isScrolling())
      {
         if (rows == 0)
         {
            return;
         }

         if (rows < 0)
         {
            backward(-rows);
            return;
         }

         prepareReposition();

         cursor.forward(rows);
         offEnd = cursor.getOffEnd();
      }
      else
      {
         super.forward(rows);
      }
   }

   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   ids
    *          Array of primary key IDs, arranged from left to right to
    *          coincide with the records being joined by the underlying
    *          query.
    * @param   forceFetch
    *          <code>true</code> to override the current fetch-on-reposition
    *          setting to <code>true</code>, in case the underlying
    *          <code>Results</code> implementation attempts to fetch the
    *          record during repositioning;  <code>false</code> to honor the
    *          current setting.
    *
    * @return  <code>true</code> if reposition was successful;
    *          <code>false</code> if unsuccessful.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   protected boolean repositionByID(Serializable[] ids, boolean forceFetch)
   {
      return repositionByID(ids, forceFetch, true);
   }

   /**
    * Reposition the cursor such that a request to retrieve the next result
    * will retrieve the result which matches the specified array of primary
    * key IDs.  A request to retrieve the previous result will retrieve the
    * result immediately previous to this, if any, in the results list.
    *
    * @param   ids
    *          Array of primary key IDs, arranged from left to right to
    *          coincide with the records being joined by the underlying
    *          query.
    * @param   forceFetch
    *          <code>true</code> to override the current fetch-on-reposition
    *          setting to <code>true</code>, in case the underlying
    *          <code>Results</code> implementation attempts to fetch the
    *          record during repositioning;  <code>false</code> to honor the
    *          current setting.
    * @param   scan
    *          For scrolling queries: <code>true</code> to scan all results in
    *          the event a match is not found in the cache;  <code>false</code>
    *          to check the cache only.
    *
    * @return  <code>true</code> if reposition was successful;
    *          <code>false</code> if unsuccessful.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   protected boolean repositionByID(Serializable[] ids, boolean forceFetch, boolean scan)
   {
      if (isScrolling())
      {
         prepareReposition();
         boolean done = false;
         try
         {
            done = cursor.reposition(ids, scan);
         }
         finally
         {
            offEnd = cursor.getOffEnd();
            afterReposition(done);
         }

         return done;
      }
      else
      {
         return super.repositionByID(ids, forceFetch);
      }
   }

   /**
    * Return the number of results found for this query.
    * <p>
    * Note that the first time this method is invoked, the database server
    * may have to scroll to the end of the result set, which may be a
    * time-consuming process for a large result set.
    *
    * @return  Size of result set or value if the query is not open.
    */
   public integer size()
   {
      if (isScrolling())
      {
         return cursor.size();
      }
      else
      {
         return super.size();
      }
   }

   /**
    * Retrieve the next composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void next(LockType lockType)
   {
      if (isScrolling() && cursor.isFoundFirst())
      {
         Object[] data;

         do
         {
            data = cursor.getNext();

            if (data == null && (cursor.isFullSet() || !cursor.isForwardScroll()))
            {
               offEnd = OffEnd.BACK;

               try
               {
                  fetch(false, lockType, false);
               }
               catch (MissingRecordException e)
               {
                  // cannot happen, we are not loading a record
               }

               return;
            }

            if (data != null)
            {
               if (loadByValues(data, lockType))
                  return;
            }

         } while (data != null);
      }

      boolean connected = (adapter != null && adapter.isConnected());
      Results results = getResults();
      
      boolean missing;
      do
      {
         missing = false;
         try
         {
            boolean avail;
            if (betweenRows && connected && !isScrolling())
            {
               avail = true;
            }
            else
            {
               errorCondition = false;
               if (!results.next())
               {
                  if (errorCondition)
                  {
                     // if an error condition was detected (and handled) then
                     // we should retry next() action
                     next(lockType);
                     
                     return;
                  }
                  else
                  {
                     avail = false;
                  }
               }
               else
               {
                  avail = true;
               }
            }
            
            offEnd = (avail ? OffEnd.NONE : OffEnd.BACK);
            fetch(avail, lockType, false);

            if (isScrolling())
            {
               if (avail)
               {
                  cursor.addResultNext(getCurrentIds());
               }
               else
               {
                  cursor.addResultNext(null);
               }
            }
         }
         catch (QueryOffEndException exc)
         {
            if (isScrolling())
            {
               cursor.addResultNext(null);
            }

            offEnd = OffEnd.BACK;
            if (!isLenientOffEnd())
            {
               exc.fillInStackTrace();
               throw exc;
            }
         }
         catch (MissingRecordException exc)
         {
            missing = true;
         }
         finally
         {
            betweenRows = false;
         }
      } while (missing);
   }

   /**
    * Fetch the array of primary key IDs associated with the next result in
    * this query's result list.
    *
    * @return  Array of record IDs reflecting the next query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekNext()
   {
      Object[] res = super.peekNext();
      if (isScrolling())
         cursor.addResultNext(res);
      return res;
   }

   /**
    * Fetch the array of primary key IDs associated with the previous result
    * in this query's result list.
    *
    * @return  Array of record IDs reflecting the previous query result, or
    *          <code>null</code> if there is no such result.
    */
   public Object[] peekPrevious()
   {
      Object[] res = super.peekPrevious();
      if (isScrolling())
         cursor.addResultPrevious(res);
      return res;
   }

   /**
    * Retrieve the previous composite row of results for the query.  The
    * underlying buffer for each query component is updated with the
    * appropriate record.
    *
    * @param   lockType
    *          Lock type to apply to records retrieved (overrides default
    *          lock type set for each query component).
    *
    * @throws  ErrorConditionException
    *          if the query or the fetch of any record fails, or if no records
    *          are available and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   public void previous(LockType lockType)
   {
      if (isScrolling() && cursor.isFoundFirst())
      {
         Object[] data;

         do
         {
            data = cursor.getPrevious();

            if (data == null && (cursor.isFullSet() || cursor.isForwardScroll()))
            {
               offEnd = OffEnd.FRONT;

               try
               {
                  fetch(false, lockType, false);
               }
               catch (MissingRecordException e)
               {
                  // cannot happen, we are not loading a record
               }

               return;
            }

            if (data != null)
            {
               if (loadByValues(data, lockType))
                  return;
            }

         } while (data != null);
      }

      boolean setToLast = treatPreviousAsLast();

      try
      {
         // When moving backward, all revalidation bets are off.
         breakValue = null;
         
         super.previous(lockType);

         if (isScrolling())
         {
            if (!isOffEnd().booleanValue())
            {
               cursor.addResultPrevious(getCurrentIds());
            }
            else
            {
               cursor.addResultPrevious(null);
            }
         }
         
         return;
      }
      catch (UnsupportedOperationException exc)
      {
         // We tried to do something the underlying Results implementation
         // could not do;  invalidate and try again.
         if (isPreselect())
         {
            invalidate();
            if (!setToLast)
            {
               execute();
            }
            
            super.previous(lockType);

            if (isScrolling())
            {
               if (!isOffEnd().booleanValue())
               {
                  cursor.addResultPrevious(getCurrentIds());
               }
               {
                  cursor.addResultPrevious(null);
               }
            }
         }
         else
         {
            throw exc;
         }
      }
   }
   
   /**
    * Move the current cursor position backward by the specified number of
    * rows.  This will always leave the cursor between two results.  For
    * example, given a request to move 1 row backward:
    * <ul>
    *   <li>if the cursor currently is positioned <i>directly on a result</i>,
    *       this would move it back <i>one half position</i>, to just
    *       before the current result;
    *   <li>if the cursor currently is positioned <i>between two results</i>,
    *       this would move it back <i>one position</i>, to just before the
    *       current result.
    * </ul>
    *
    * @param   rows
    *          Number of rows to scroll the cursor backward.
    *
    * @throws  ErrorConditionException
    *          if there is an error repositioning the cursor.
    */
   public void backward(int rows)
   {
      if (isScrolling())
      {
         if (rows == 0)
         {
            return;
         }

         if (rows < 0)
         {
            forward(-rows);
            return;
         }

         prepareReposition();

         cursor.backward(rows);
         offEnd = cursor.getOffEnd();
      }
      else
      {
         if (rows != 0)
         {
            // When moving backward, all revalidation bets are off.
            breakValue = null;
         }

         super.backward(rows);
      }
   }

   /**
    * Reset the query by clearing its state to a known default.  This is
    * used to ensure that subsequent requests for FIRST and LAST record
    * retrievals operate on a known query/buffer state.
    * <p>
    * This implementation nulls out the current result set, so that the next
    * record retrieval request triggers a new server-side query.  Optionally,
    * it also resets resolvable substitution arguments and triggers where
    * expression arguments to be resolved, if a client-side where cause
    * expression is defined.
    * <p>
    * For scrolling queries cache is reset.
    *
    * @param   resolveArgs
    *          <code>true</code> if query substitution arguments should be
    *          resolved/reset as part of the reset;  <code>false</code> if
    *          existing argument values should be used for the next query
    *          execution.
    */
   public void reset(boolean resolveArgs)
   {
      super.reset(resolveArgs);

      if (isScrolling())
      {
         cursor.reset();
      }

      invalidComponent = null;
   }
   
   /**
    * Load one or more records into their associated buffers, given an array
    * of primary key ID values, or the actual DMOs themselves.  The IDs/DMOs
    * are arranged to coincide with the tables underlying the query, from
    * left-most to right-most (in the sense of how they are joined by the
    * query).
    * <p>
    * In the event an expected value cannot be loaded (e.g., the record has
    * been deleted or is otherwise no longer available), this method throws
    * {@link MissingRecordException}, after setting the associated buffer(s)
    * into {@link RecordBuffer#setUnknownMode() unknown mode}. Note that a
    * record being locked by another session does not constitute a "missing"
    * record. Note also that this exception is not thrown until all the
    * elements of <code>data</code> have been processed to the extent
    * possible.
    *
    * @param   data
    *          Array of primary key IDs or DMOs, one per table involved in the
    *          query.
    * @param   lockType
    *          Lock type which should by used.  Set to <code>null</code> to
    *          allow query to use its current lock type.
    * 
    * @throws  MissingRecordException
    *          if any record cannot be loaded, because it is no longer
    *          available (e.g., deleted, etc.).
    * @throws  PersistenceException
    *          if there is an error loading data.
    */
   public void load(Object[] data, LockType lockType)
   throws MissingRecordException,
          PersistenceException
   {
      if (isScrolling())
      {
         Serializable[] arr = new Serializable[data.length];
         System.arraycopy(data, 0, arr, 0, data.length);
         if (repositionByID(arr, true, false))
         {
            cursor.next();
            offEnd = OffEnd.NONE;
         }
      }
      else if (isPreselect())
      {
         super.load(data, lockType);
      }

      try
      {
         coreFetch(data, lockType, false);
      }
      catch (MissingRecordException exc)
      {
         setEmptyBuffersUnknown();

         throw exc;
      }
   }
   
   /**
    * Assemble an array of primary key IDs for the current record(s) in the
    * buffers underlying this query, from left-most to right-most (in the
    * sense of how the query joins the associated tables).  This method
    * essentially takes a lightweight snapshot of the currently loaded
    * record(s), so that this may be restored later.
    * <p>
    * This method overrides the parent's implementation to handle the case
    * where we have just transitioned from preselect to dynamic mode, but we
    * do not yet have a <code>Results</code> object from which to gather data.
    * We generate a dynamic <code>Results</code> object and position to the
    * first result, then use the parent's implementation to gather the data.
    *
    * @return  Array of primary key IDs or DMOs, one per table involved in the
    *          query.
    *
    * @see     #load
    */
   public Object[] getRow()
   {
      Object[] srcData = super.getRow();
      if (srcData == null && !isPreselect())
      {
         Results results = getResults();
         results.first();
         srcData = super.getRow();
      }

      return srcData;
   }

   /**
    * Assemble an array of primary key IDs for the current record(s) in the buffers underlying
    * this query, from left-most to right-most (in the sense of how the query joins the associated
    * tables). This method essentially takes a lightweight snapshot of the currently loaded
    * record(s), so that this may be restored later.<p>
    * For scrolling queries it gets the row from the cache instead of interrogating
    * {@link #results}. For non-scrolling queries it works in the same way as {@link #getRow()}.
    *
    * @return  Array of primary key IDs, one per table involved in the query.
    *
    * @see     #load(Object[], LockType)
    */
   public Object[] getCachedRow()
   {
      if (isScrolling())
      {
         return cursor.getRow();
      }
      else
      {
         return getRow();
      }
   }
   
   /**
    * Respond to a dirty (i.e., error-driven) session closing event by
    * invalidating any current, preselected results.  This will cause a
    * switch to dynamic mode, if the query is not associated with a browse.
    * If we are in dynamic mode already for a clean session close, this
    * method is a no-op.
    * <p>
    * On a clean (i.e., no error) session closing event, delegate the event
    * to the results adapter.
    *
    * @param   event
    *          Session event type.
    *
    * @return  Under normal conditions, <code>false</code>, indicating this
    *          query should not be deregistered as a session listener after
    *          this method returns.  Under an error condition, return
    *          <code>true</code>.
    *
    * @throws  PersistenceException
    *          not thrown (required by interface).
    */
   public boolean sessionEvent(SessionListener.Event event)
   throws PersistenceException
   {
      boolean deregister =
         SessionListener.Event.SESSION_CLOSING_WITH_ERROR.equals(event);
      
      if (closed)
      {
         return deregister;
      }
      
      switch (event)
      {
         case TRANSACTION_COMMITTING:
            
            if (cacheOnReval && isPreselect())
            {
               if (adapter.isConnected())
               {
                  Results old = adapter.getResults();
                  SimpleResults cached = cacheResults(old);
                  if (old != cached)
                  {
                     old.sessionClosing();
                  }
                  adapter.setResults(cached);
               }
            }
            else
            {
               adapter.sessionClosing();
            }
            
            resultsCached = true;
            
            break;
            
         case SESSION_CLOSING_NORMALLY:
            if (!resultsCached)
            {
               resetResults();
            }
            break;
            
         case SESSION_CLOSING_WITH_ERROR:
            invalidate(invalidComponent);
            break;
            
         default:
            return false;
      }
      
      return deregister;
   }
   
   /**
    * Create a string representation of this object's state, primarily for
    * debug purposes.
    * 
    * @return  String representation of this object.
    */
   public String toString()
   {
      StringBuilder bld = new StringBuilder();
      bld.append("[");
      bld.append(isPreselect() ? "PRE" : "DYN");
      bld.append("] ");
      bld.append(super.toString());

      return bld.toString();      
   }
   
   /**
    * Indicate whether this query preselects all of its results.  This
    * implementation initially preselects its results;  however, the current
    * result set may be invalidated, at which point the query switches to a
    * dynamic retrieval mode.
    *
    * @return  <code>true</code> if the query currently is in preselect mode,
    *          <code>false</code> if its result set has been invalidated.
    */
   public boolean isPreselect()
   {
      return (invalidComponent == null);
   }

   /**
    * Report the record buffers for whose changes this object is interested
    * in listening.  This listener will be registered with the {@link
    * ChangeBroker} to receive notifications of any change to DMOs whose
    * interface types match those returned by the {@link
    * RecordBuffer#getDMOInterface} methods of the returned buffers.
    *
    * @return  An iterator on the record buffers stored in each of this
    *          query's components.
    */
   public Iterator<RecordBuffer> recordBuffers()
   {
      return Arrays.asList(getRecordBuffers()).iterator();
   }
   
   /**
    * Respond to a record change event by invalidating the current preselect
    * results, if currently in preselect mode and the state change impacts
    * the validity of the current results.
    *
    * @param   event
    *          Event which describes the DMO state change.
    *
    * @throws  PersistenceException
    *          if any error occurs while processing the state change event.
    */
   public void stateChanged(RecordChangeEvent event)
   throws PersistenceException
   {
      boolean preselect = isPreselect();
      boolean invalidate = true;
      boolean cacheAll = false;
      
      if (adapter.isConnected())
      {
         setupInvalidationProperties();
         
         if (event.isDelete())
         {
            // TODO:  make this check more specific and more restrictive.
            // Specifically, we only want to invalidate if we know the
            // deleted record is included in our preselected results (not
            // including the record currently being visited).
            
            // Optimization:  if the source of the event is a buffer used by
            // this query, do not go into dynamic mode.  Do not perform this
            // optimization for scrolling queries, which can navigate back to
            // the deleted record.
            if (!isScrolling())
            {
               Iterator<Component> iter = components();
               while (invalidate && iter.hasNext())
               {
                  Component c = iter.next();
                  RecordBuffer buffer = c.getBuffer();
                  
                  if (event instanceof GlobalChangeEvent)
                  {
                     GlobalChangeEvent gce = (GlobalChangeEvent) event;
                     if (gce.getEntity().equals(buffer.getEntityName()))
                     {
                        Persistable bDMO = buffer.getCurrentRecord();
                        Serializable pk = gce.getPrimaryKey();
                        if (bDMO != null && !bDMO.getId().equals(pk))
                        {
                           continue;
                        }
                     }
                     else
                     {
                        continue;
                     }
                  }
                  else if (buffer != event.getSource())
                  {
                     continue;
                  }
                  
                  if (cacheOnReval)
                  {
                     invalidate = false;
                  }
                  
                  cacheOnReval = isPreselect();
                  
                  // The cacheAll optimization is meant to handle the case of
                  // a delete loop.  If the first record we visited in a
                  // progressive result set is deleted by our backing buffer,
                  // we assume a primary purpose of the current loop is to
                  // delete the records found by the query.  This is not
                  // guaranteed, but it is the common case for the
                  // circumstances we test for below.  The idea is to let the
                  // invalidation proceed, but then immediately to re-find all
                  // of the remaining records by re-executing the query.
                  if (invalidate &&
                      cacheOnReval &&
                      (adapter.getResults() instanceof ProgressiveResults) &&
                      adapter.getRowNumber() == 0)
                  {
                     cacheAll = true;
                  }
               }
            }
         }
         else if (event.isInsert())
         {
            if (preselect || breakValue != null)
            {
               invalidate = doubleCheckInvalidation(event);
            }
            if (invalidate)
            {
               cacheOnReval = false;
            }
         }
         else
         {
            if (preselect || breakValue != null)
            {
               // Quick out:  AdaptiveQuery sorts on an index only, which
               // cannot contain extent fields.  If the event is about extent
               // field changes only, we do not need to invalidate.
               if (event.isIndexedPropertyChangeOnly())
               {
                  invalidate = false;
               }
               else
               {
                  // This is the update case.  If any property changed is one
                  // of the properties used by this query to determine the
                  // content or sort order of the result set, we invalidate.
                  // Otherwise, do nothing.
                  String entity = event.getEntity();
                  Set<String> indexProps = invalidationProperties.get(entity);
                  Set<String> changedProps = event.getProperties().keySet();
                  
                  // Sanity check.
                  if (indexProps == null || changedProps == null)
                  {
                     return;
                  }
                  
                  // Test each changed property to see if it is a member of
                  // the set of index properties for the modified DMO type.
                  invalidate = false;
                  Iterator<String> iter = changedProps.iterator();
                  while (iter.hasNext())
                  {
                     if (indexProps.contains(iter.next()))
                     {
                        invalidate = doubleCheckInvalidation(event);
                        
                        break;
                     }
                  }
               }
            }
            
            if (invalidate)
            {
               cacheOnReval = false;
            }
         }
         
         if (invalidate)
         {
            if (preselect)
            {
               invalidate(event);
            }
            else
            {
               resetDynamicScrolling();
            }
         }
         
         if (LOG.isDebugEnabled())
         {
            if (invalidate)
            {
               LOG.debug("ALLOWED invalidation " + this);
            }
            else
            {
               LOG.debug("AVOIDED invalidation " + this);
            }
         }
         
         // The first record found by this query was deleted.  Switch back
         // into preselect mode immediately, and re-find all (remaining)
         // matching records with a scrolling query that will not be subject
         // to the "dropped" records problem encountered when using
         // ProgressiveResults within a delete loop.
         if (cacheAll)
         {
            if (LOG.isDebugEnabled())
            {
               LOG.debug("Forcing full cache of results after delete event " +
                         this);
            }
            
            // Allows execute() to determine that we now are operating in
            // preselect mode.
            invalidComponent = null;
            
            // Allows super.execute() to determine that we need a new
            // preselect result set.
            setResults(null);
            
            // Retrieve preselect result set.  Adapter will be reconnected to
            // new delegate results.
            execute();
         }
      }
      
      // Prevent a transition from dynamic to preselect mode.
      if (invalidate && !preselect && breakValue != null)
      {
         breakValue = null;
         
         if (LOG.isTraceEnabled())
         {
            LOG.trace("Canceling transition from dynamic to preselect mode");
         }
      }
   }
   
   /**
    * This function is called when ProgressiveResults has detected changes in
    * retrieved results. It switches mode from preselect to dynamic by
    * calling invalidate() using as invalid the first AdaptiveComponent. This
    * approach works while AdaptiveQuery does not support multi-table queries.
    */
   public void resultsChanged()
   {
      errorCondition = true;
      invalidate();
   }
   
   /**
    * Clean up query resources when this query's scope ends.  This
    * implementation deregisters for global change notifications, if
    * necessary, then delegates to the parent's implementation.
    * 
    * @see  PreselectQuery#cleanup()
    */
   public void cleanup()
   {
      if (closed)
      {
         return;
      }

      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         AdaptiveComponent next = (AdaptiveComponent) iter.next();
         next.cleanup();
      }
      
      super.cleanup();
   }

   /**
    * Makes this query scrolling. This method must be invoked before any results are retrieved.
    */
   public void setScrolling()
   {
      super.setScrolling();
      cursor = new Cursor(this);
   }

   /**
    * Indicate whether this query is a preselect query by its nature.
    *
    * @return  <code>false</code>.
    */
   public boolean isNativelyPreselect()
   {
      return false;
   }

   /**
    * Deletes the current row of a query’s result list. Implements the 4GL
    * DELETE-RESULT-LIST-ENTRY() method.
    *
    * @param  allowBetweenRows
    *         If <code>true</code> then the cursor can be positioned between rows (the next
    *         row will be deleted). <code>false</code> for conventional DELETE-RESULT-LIST-ENTRY()
    *         mode where the cursor should be positioned on a row (otherwise <code>false</code> is
    *         returned).
    *
    * @return <code>true</code> on success.
    */
   public logical deleteResultListEntry(boolean allowBetweenRows)
   {
      if (isScrolling())
      {
         boolean res = cursor.deleteResultListEntry(allowBetweenRows);
         if (res)
         {
            currentRowDeleted = true;
         }
         return new logical(res);
      }
      else if (!currentRowDeleted && foundFirst)
      {
         currentRowDeleted = true;
         return  new logical(true);
      }

      return new logical(false);
   }
   
   /**
    * Indicate whether this query type is scrolling by its nature.
    * 
    * @return  <code>false</code>.
    */
   protected boolean isNativelyScrolling()
   {
      return false;
   }

   /**
    * Report whether this query must check for changes across all sessions and
    * possibly invalidate its current results if any are found.  If not, this
    * query will only honor local changes to invalidate results.
    * 
    * @return  <code>true</code> to require global invalidation checks;
    *          <code>false</code> to only check locally for changes which may
    *          invalidate results.
    */
   protected boolean isGlobalInvalidation()
   {
      return globalInvalidation;
   }
   
   /**
    * Assemble the restriction (i.e., where) clause for an HQL query. If more
    * than one component contains a where clause, they are combined with a
    * logical "and". Form of the clause is:
    * <pre>
    *    where ({component0_clause})
    *     [and ({component1_clause})
    *     [...]]
    * </pre>
    * <p>
    * Please refer to the Hibernate reference documentation for permissible
    * syntax of this clause.
    *
    * @param   sortCriteria
    *          List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @return assembled where clause.
    */
   protected HQLExpression assembleWhereClause(List<SortCriterion> sortCriteria)
   throws PersistenceException
   {
      HQLExpression buf = super.assembleWhereClause(sortCriteria);
      
      // If this query was last in dynamic mode, we need to use the break
      // value provided by the DynamicResults object as a reference point at
      // which to begin our results.  Put the comparison operation in the
      // where clause with a placeholder which will be substituted with the
      // break value at execute time.
      if (breakValue != null)
      {
         SortCriterion crit = sortCriteria.get(0);
         
         if (buf.isEmpty())
         {
            buf.append("where (");
         }
         else
         {
            buf.append(" and (");
         }
         
         buf.append(crit.toWhereExpression(true, false));
         buf.append(") ");
      }

      return buf;
   }
   
   /**
    * Assemble the <code>order by</code> clause for an HQL query by combining
    * the sort criteria from each query component sequentially.  Uses the
    * form:
    * <pre>
    *   order by {{dmo.property} {asc/desc}} [, ...]
    * </pre>
    * <p>
    * Any text-based sort criteria are augmented with the necessary upper()
    * and rtrim() functions.
    *
    * @param   buf
    *          String buffer into which clause is assembled.
    *
    * @return  List of <code>SortCriterion</code> objects which describe the
    *          sort behavior for this query.
    *
    * @throws  PersistenceException
    *          if the sort clause of any query component contains any
    *          unqualified property names;  if a sort clause contains a
    *          reference to an unrecognized record buffer.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which setup data is required.
    */
   protected List<SortCriterion> assembleOrderByClause(StringBuilder buf)
   throws PersistenceException
   {
      List<SortCriterion> all = new ArrayList<SortCriterion>();
      Iterator<Component> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         AdaptiveComponent next = (AdaptiveComponent) iter.next();
         List<SortCriterion> sortCriteria = next.getSortCriteria();
         all.addAll(sortCriteria);
         
         if (i == 0)
         {
            buf.append("order by ");
         }
         else
         {
            buf.append(" ");
         }
         
         assembleOrderByClauseNoPreamble(buf, sortCriteria);
      }
      
      return all;
   }
   
   /**
    * Prepare query substitution parameters for this query.  In addition to
    * the basic parameters handled by the superclass, this method will add
    * the query's <code>breakValue</code> as a parameter if it is non-null.
    * A non-<code>null</code> <code>breakValue</code> indicates that the
    * query is shifting from dynamic mode to preselect mode.  All records
    * retrieved from this point forward must not include those already
    * retrieved, and the <code>breakValue</code> demarcates this boundary.
    * The where clause will have been modified to include a comparison
    * operation with the <code>breakValue</code> in this case.
    *
    * @param   args
    *          Argument list to be populated.
    * @param   types
    *          Type list to be populated.  This may be populated with
    *          <code>null</code> values, but it must coincide in size and in
    *          the absolute positions of its elements with the
    *          <code>args</code> list.  Any <code>null</code> element will
    *          be filled in later.
    */
   protected void prepareParameters(List<Object> args, List<Type> types)
   throws PersistenceException
   {
      super.prepareParameters(args, types);
      
      if (breakValue == null && invalidComponent != null)
      {
         Persistable dmo = invalidComponent.getBuffer().getSnapshot();
         if (dmo != null)
         {
            AdaptiveComponent comp = (AdaptiveComponent) components().next();
            List<SortCriterion> sortCriteria = comp.getSortCriteria();
            int len = sortCriteria.size();
            Method getter = null;
            for (int i = 0; i < len && getter == null; i++)
            {
               SortCriterion crit = sortCriteria.get(i);
               getter = crit.getMethod();
            }

            if (getter != null)
            {
               try
               {
                  breakValue = getter.invoke(dmo);
               }
               catch (Exception exc)
               {
                  RecordBuffer buffer = invalidComponent.getBuffer();
                  DBUtils.handleException(buffer.getDatabase(), exc);
                  LOG.error(exc);
               }
            }
         }
      }
      
      if (breakValue != null)
      {
         args.add(breakValue);
         types.add(null);
      }
   }
   
   /**
    * Overrides the parent's pure preselect implementation to use a different
    * retrieval techniques depending upon the current mode of the query.  As
    * long as the query remains in preselect mode, the superclass' approach
    * is used.  If, however, the preselect result set has been invalidated by
    * a DMO state change, this method creates engages the "backup" dynamic
    * retrieval mechanism.
    * <p>
    * In this case, a logical <code>DynamicResults</code> object is created
    * the first time through this method after invalidation.  Thereafter,
    * this object is simply used as a regular {@link Results} object, in that
    * parent's {@link PreselectQuery#fetch} method uses the new results
    * object without knowing or caring that it has been substituted with a
    * dynamic implementation.
    * <p>
    * In addition, if this query is set to check for global DMO change events,
    * such events are collected and processed, which will possibly invalidate
    * the query's current results, or cause a change to dynamic mode.
    *
    * @return  <code>true</code> if the underlying preselect result set was
    *          created during this invocation;  <code>false</code> if query
    *          was executed previously and we already have a preselect result
    *          set.
    *
    * @throws  ErrorConditionException
    *          if an error occurs executing the preselect query.
    * @throws  IllegalStateException
    *          if no query component has been added to this query.
    */
   protected boolean execute()
   {
      // Check if all components with a non-null JOIN have a valid inverse
      // record on which to perform the join
      if (!verifyJoins())
      {
         return true;
      }
      
      // Possibly invalidate if another session has made a change.
      if (isGlobalInvalidation())
      {
         Iterator<Component> iter = components();
         while (iter.hasNext())
         {
            AdaptiveComponent comp = (AdaptiveComponent) iter.next();
            comp.processGlobalEvents();
         }
      }
      
      if (isPreselect())
      {
         try
         {
            Iterator<Component> iter = components();
            while (iter.hasNext())
            {
               AdaptiveComponent comp = (AdaptiveComponent) iter.next();
               comp.initialize();
            }
            
            return super.execute();
         }
         catch (PersistenceException exc)
         {
            throw new StopConditionException(exc);
         }
         finally
         {
            breakValue = null;
         }
      }
      
      if (adapter.isConnected())
      {
         return false;
      }
      
      // If we call super.execute() above, a cleaner is registered for us,
      // but if not, we have to do it here.
      registerCleaner();
      
      try
      {
         // Create a dynamic results object which will drive record retrieval.
         DynamicQuery query = null;
         int tc = getTableCount();
         if (tc > 1)
         {
            query = createCompoundQuery(invalidComponent);
         }
         else
         {
            query = createSimpleQuery(invalidComponent);
         }
         
         boolean idOnly[] = new boolean[tc];
         Iterator<Component> iter = components();
         for (int i = 0; iter.hasNext(); i++)
         {
            Component component = iter.next();
            idOnly[i] = !isClientWhere() && component.isIDOnly();
         }
         DynamicResults dr = new DynamicResults(query, idOnly);
         Results hlr = createResults(dr);
         setResults(hlr);
      }
      catch (PersistenceException exc)
      {
         ErrorManager.recordOrThrowError(exc);
      }
      
      return false;
   }
   
   /**
    * Respond to the condition that a component with a non-null JOIN has a
    * missing inverse on which to perform the join.
    * <p>
    * This implementation is a placeholder which currently defers to the
    * parent's implementation.
    */
   protected void handleMissingJoinRecord()
   {
      /*
      if (!isPreselect() || revalCount > 0)
      {
         throw new QueryOffEndException("Missing join record");
      }
      */
      
      super.handleMissingJoinRecord();
   }

   /**
    * Indicate whether a call to move to the previous record in the result list should be treated
    * as a call to move to the last record in the list. This will be the cases listed in the super
    * function {@link PreselectQuery#treatPreviousAsLast}, but only if the query is in preselect
    * mode. That fixes previous mistakenly treated as last after query invalidation.
    *
    * @return  <code>true</code> to treat a previous request as a request to move to the last
    *          record in the list;  else <code>false</code>.
    */
   protected boolean treatPreviousAsLast()
   {
      return isPreselect() && super.treatPreviousAsLast();
   }
   
   /**
    * Execute the query to create a first level result set.  This
    * implementation creates an instance of <code>ProgressiveResults</code>,
    * which defers actually executing the query until the first request to
    * fetch a result.
    *
    * @param   persistence
    *          Object which is used to execute the query.
    * @param   hql
    *          HQL query string.
    * @param   args
    *          Substitution parameters to be inserted into placeholders within
    *          within the HQL query string.
    * @param   types
    *          Query substitution parameter types.
    *
    * @return  An instance of <code>ProgressiveResults</code> which fetches
    *          progressively larger brackets of query results as the results
    *          are accessed.
    *
    * @throws  PersistenceException
    *          not thrown by this implementation.
    */
   protected Results executeQuery(Persistence persistence,
                                  String hql,
                                  Object[] args,
                                  Type[] types)
   throws PersistenceException
   {
      Results results = null;
      
      if (cacheOnReval)
      {
         results = super.executeQuery(persistence, hql, args, types);
      }
      else
      {
         ProgressiveResults pr = new ProgressiveResults(persistence,
                                                        hql,
                                                        args,
                                                        types,
                                                        isScrolling());
         pr.addResultsChangeListener(this);
         results = pr;
      }
      
      return results;
   }
   
   /**
    * Override the parent's implementation to intercept result creation, and
    * always return our adapter, after setting the <code>results</code>
    * parameter as the adapter's delegate.
    *
    * @param   results
    *          An object which manipulates the database result set returned
    *          from execution of the query.
    *
    * @return  <code>Results</code> object.
    *
    * @throws  PersistenceException
    *          if there is any error creating results.
    */
   protected Results createResults(Results results)
   throws PersistenceException
   {
      adapter.setResults(results);
      
      betweenRows = false;
      
      if (highLevelResults == null)
      {
         highLevelResults = super.createResults(adapter);
      }
      
      return highLevelResults;
   }
   
   /**
    * Create a single-table, dynamic query to be used as the worker object
    * for a <code>DynamicResults</code> object when this query switches from
    * preselect to dynamic mode.  Creates a modified implementation of
    * <code>RandomAccessQuery</code>, which does not reset the record buffer
    * upon instantiation.
    *
    * @param   invalidComp
    *          Query component which caused this query to be invalidated.
    *
    * @return  A <code>RandomAccessQuery</code> based upon this query's single
    *          component.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which setup data is required.
    */
   protected DynamicQuery createSimpleQuery(AdaptiveComponent invalidComp)
   throws PersistenceException
   {
      // Always do the initial fetch with lock type of NONE.  It will be
      // overridden as necessary in PreselectQuery.fetch().
      LockType lockType = LockType.NONE;
      
      AdaptiveComponent comp = (AdaptiveComponent) components().next();
      
      return (new RandomAccessQuery(comp.getBuffer().getDMOProxy(),
                                    comp.getOriginalWhere(),
                                    null,
                                    comp.getOriginalSort(),
                                    comp.getInverse(),
                                    comp.getRawArgs(),
                                    lockType)
         {
            protected void initializeBuffer() {};
         });
   }
   
   /**
    * Create a multi-table, compound query to be used as the worker object
    * for a <code>DynamicResults</code> object when this query switches from
    * preselect to dynamic mode.  Uses the parent query's components to
    * define a <code>Joinable</code> instance for each of the compound
    * query's own components.
    * <p>
    * Single-table <code>AdaptiveQuery</code> instances are used for every
    * joinable component, except the one corresponding to the buffer which
    * caused invalidation of the preselect result set.  For the component
    * corresponding with this buffer, an instance of
    * <code>RandomAccessQuery</code> is used instead.
    *
    * @param   invalidComp
    *          Query component which caused this query to be invalidated.
    *
    * @return  A <code>CompoundQuery</code> based upon this query's multiple
    *          components.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which setup data is required.
    */
   protected DynamicQuery createCompoundQuery(AdaptiveComponent invalidComp)
   throws PersistenceException
   {
      // Create a CompoundQuery, resolve once, not preselect.
      CompoundQuery query = new CompoundQuery(true, false);
      
      Class<?> invalidType = invalidComp.getBuffer().getDMOInterface();
      Iterator<Component> iter = components();
      for (int i = 0; iter.hasNext(); i++)
      {
         AdaptiveComponent comp = (AdaptiveComponent) iter.next();
         Joinable joinable = null;
         RecordBuffer buffer = comp.getBuffer();
         Class<?> type = buffer.getDMOInterface();
         
         if (type.equals(invalidType))
         {
            joinable = new RandomAccessQuery(buffer.getDMOProxy(),
                                             comp.getOriginalWhere(),
                                             null,
                                             comp.getOriginalSort(),
                                             comp.getInverse(),
                                             comp.getRawArgs(),
                                             comp.getLockType());
         }
         else
         {
            joinable = new AdaptiveQuery(buffer.getDMOProxy(),
                                         comp.getOriginalWhere(),
                                         null,
                                         comp.getOriginalSort(),
                                         comp.getInverse(),
                                         comp.getRawArgs(),
                                         comp.getLockType());
         }
         
         query.addComponent(joinable);
      }
      
      return query;
   }
   
   /**
    * Invalidate the preselected result set of this query, which effectively
    * switches this query from preselect mode to dynamic mode.
    *
    * @param   invalidComponent
    *          Query component which is triggering invalidation.
    */
   protected void invalidate(AdaptiveComponent invalidComponent)
   {
      if (!isPreselect())
      {
         // If we are already in dynamic mode, invalidation means clearing the
         // dynamic query's cursor cache.  This prevents us from getting a
         // potentially stale value for first() and last() navigation.
         resetDynamicScrolling();
      }
      
      this.invalidComponent = invalidComponent;
      
      if (adapter.isConnected())
      {
         resetResults();
      }
      else
      {
         clearRepoCache();
      }
      
      adapter.prepareDelegateChange();
   }
   
   /**
    * Reset the state of the results adapter to its initial default state and
    * invoke the superclass' overridden method.
    */
   protected void resetResults()
   {
      resultsCached = false;

      super.resetResults();
      
      adapter.cleanup();
      adapter.setResults(null);
   }

   /**
    * Get array of the primary keys of the records contained in the backing buffers.
    *
    * @return  array of the primary keys of the records contained in the backing buffers or
    *          <code>null</code> if any of the buffers doesn't contain a record.
    */
   protected Object[] getCurrentIds()
   {
      List<Long> data = new ArrayList<>();

      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         Component comp = iter.next();

         RecordBuffer buffer = comp.getBuffer();
         if (!buffer.isAvailable())
         {
            return null;
         }

         data.add(buffer.rowID().getValue());
      }

      return data.toArray();
   }

   /**
    * Load all records for a given result row into their backing buffers.
    *
    * @param  data
    *         An array of primary key IDs or DMOs representing records to be fetched for the
    *         composite row.
    * @param  lockType
    *         Overriding lock type to apply to fetched records. If null, default lock types are
    *         applied.
    *
    * @return <code>true</code> if the record(s) were successfully loaded.
    */
   protected boolean loadByValues(Object[] data, LockType lockType)
   {
      currentRowDeleted = false;

      if (data == null)
         return false;

      try
      {
         offEnd = OffEnd.NONE; // required for PreselectQuery to fetch properly
         coreFetch(data, lockType, true);
         accumulate();
      }
      catch (MissingRecordException e)
      {
         return false;
      }
      catch (PersistenceException e)
      {
         ErrorManager.recordOrThrowError(e);
      }

      return true;
   }

   /**
    * Fetch all records for a given result row into their backing buffers.
    * Each result row contains one column per query component.  Each column
    * contains the unique ID of the record to be fetched or the record itself.
    * If the latter, the record is refetched in case it has changed since the
    * record was first retrieved.  If the scrollable cursor backing this query
    * currently is at a position where no retrieval is possible, the backing
    * buffers will have their current records released.
    *
    * @param   available
    *          <code>true</code> if the most recent scrollable cursor move
    *          left the cursor at an available result row;  <code>false</code>
    *          if the cursor cannot read a result at its current position.
    * @param   lockType
    *          Overriding lock type to apply to fetched records.  If
    *          <code>null</code>, default lock types are applied.
    * @param   errorIfNull
    *          <code>true</code> if failed retrievals should raise error.
    *
    * @throws  MissingRecordException
    *          if the record was expected to be available, but wasn't.  This
    *          typically indicates the record was deleted between the time it
    *          initially was found by the query and the time it was to be
    *          fetched to store in the record buffer.
    * @throws  ErrorConditionException
    *          if the fetch of any record fails or if no records are available
    *          and the query is set to fail in this case.
    * @throws  QueryOffEndException
    *          if no records are available and the query is set not to fail
    *          in this case.
    */
   protected void fetch(boolean available,
                        LockType lockType,
                        boolean errorIfNull)
   throws MissingRecordException
   {
      super.fetch(available, lockType, errorIfNull);

      if (!isScrolling() &&
          !foundFirst &&
          available &&
          offEnd == OffEnd.NONE)
      {
         foundFirst = true;
      }
   }

   /**
    * Given a record change event for which this query is listening, determine
    * whether invalidation may be avoided.  For an update, assume it already
    * has been determined that a relevant property (one which can affect the
    * changed DMO's inclusion in preselected results) has been modified.
    * <p>
    * This additional check is about detecting whether it is possible that a
    * modified or newly inserted record can possibly appear later in the
    * preselected result set than the record we currently are visiting.
    * 
    * @param   event
    *          Event which describes the DMO state change.
    * 
    * @return  <code>true</code> if invalidation is confirmed;
    *          <code>false</code> to abandon invalidation.
    */
   private boolean doubleCheckInvalidation(RecordChangeEvent event)
   {
      if (isScrolling())
      {
         return true;
      }
      
      // TODO:  make this check more specific and more restrictive.
      // Specifically, we only want to invalidate if we know the inserted or
      // updated record would have been included in our preselected results.
      
      boolean invalidate = true;
      boolean isInsert = event.isInsert();
      RecordBuffer sourceBuf = event.getSource();
      Persistable targetDMO = event.getDMO();
      
      Iterator<Component> comps = components();
      while (invalidate && comps.hasNext())
      {
         AdaptiveComponent c = (AdaptiveComponent) comps.next();
         RecordBuffer buffer = c.getBuffer();
         
         if (isInsert)
         {
            if (buffer != sourceBuf)
            {
               continue;
            }
         }
         else
         {
            Persistable current = buffer.getCurrentRecord();
            if (current != targetDMO)
            {
               continue;
            }
         }
         
         Persistable snapshot = buffer.getSnapshot();
         DMOSorter sorter = c.getDMOSorter();
         if (sorter == null)
         {
            // Assume we are using default index (by primary key).  In this
            // case, an insert causes invalidation, an update does not.
            invalidate = isInsert;
         }
         else
         {
            // Use sorter to determine whether changed record *could* appear
            // further on in the results.  If not, do not invalidate now.
            int comp = sorter.compare(targetDMO, snapshot);
            if (comp <= 0)
            {
               invalidate = false;
            }
         }
      }
      
      return invalidate;
   }
   
   /**
    * If in dynamic mode, reset the cursor underlying the dynamic query in
    * order to prevent stale data.
    */
   private void resetDynamicScrolling()
   {
      if (adapter.isConnected())
      {
         Results res = adapter.getResults();
         if (res instanceof DynamicResults)
         {
            ((DynamicResults) res).query.resetScrolling();
         }
      }
   }
   
   /**
    * Set up the map of DMO implementation classes to sets of properties
    * which, if changed, would invalidate the current, preselected result set
    * for this query.  These will be properties which are used either to
    * restrict (i.e., where clause) or sort (i.e., order by clause) the
    * contents of the result set.
    * 
    * @throws  PersistenceException
    *          if there is an error creating the <code>HQLPreprocessor</code>
    *          from which setup data is required.
    */
   private void setupInvalidationProperties()
   throws PersistenceException
   {
      if (invalidationProperties != null)
      {
         return;
      }
      
      invalidationProperties = new HashMap<String, Set<String>>();
      
      Iterator<Component> iter = components();
      while (iter.hasNext())
      {
         AdaptiveComponent comp = (AdaptiveComponent) iter.next();
         
         // Collect property names from where clause.
         Map<String, Set<String>> whereProps =
            comp.getRestrictionProperties();
         if (whereProps != null)
         {
            Iterator<Map.Entry<String, Set<String>>> iter2 =
               whereProps.entrySet().iterator();
            while (iter2.hasNext())
            {
               Map.Entry<String, Set<String>> next = iter2.next();
               String entity = next.getKey();
               Set<String> props = next.getValue();
               Set<String> indexProps = invalidationProperties.get(entity);
               if (indexProps == null)
               {
                  indexProps = new HashSet<String>();
                  invalidationProperties.put(entity, indexProps);
               }
               indexProps.addAll(props);
            }
         }
         
         // Collect property names from sort criteria.
         List<SortCriterion> sortCriteria = comp.getSortCriteria();
         if (sortCriteria == null || sortCriteria.isEmpty())
         {
            continue;
         }
         
         Iterator<SortCriterion> iter3 = sortCriteria.iterator();
         while (iter3.hasNext())
         {
            SortCriterion next = iter3.next();
            Class<?> dmoClass = next.getDMOClass();
            String entity = dmoClass.getName();
            Set<String> indexProps = invalidationProperties.get(entity);
            if (indexProps == null)
            {
               indexProps = new HashSet<String>();
               invalidationProperties.put(entity, indexProps);
            }
            indexProps.add(next.getUnqualifiedName());
         }
      }
   }
   
   /**
    * Invalidate the preselected result set of this query, which effectively
    * switches this query from preselect mode to dynamic mode.
    *
    * @param   event
    *          Object which holds information about the DMO state change
    *          which occurred.
    */
   private void invalidate(RecordChangeEvent event)
   {
      String entity = event.getEntity();
      AdaptiveComponent comp = null;
      Iterator<Component> iter = components();
      while (iter.hasNext() && comp == null)
      {
         AdaptiveComponent next = (AdaptiveComponent) iter.next();
         if (next.getBuffer().getEntityName().equals(entity))
         {
            comp = next;
         }
      }
      
      assert comp != null;
      
      invalidate(comp);
      
      if (LOG.isTraceEnabled())
      {
         String hql = getHQL();
         
         if (hql != null)
            hql = hql.replaceAll("\t\n\r", "\\\\t\\\\n\\\\r");
         
         Exception exc = new Exception("Details (not an error):  [" + 
                                       hql + "]");
                                       
         LOG.trace("Adaptive query's result set invalidated by:  " +
                   invalidComponent.getBuffer().toString(event.getDMO()) +
                   (event.isDelete()
                   ? "Record deleted"
                   : "Affected properties:  " + event.getProperties()),
                   exc);
      }
   }
   
   /**
    * Invalidate the query by using the first query component as the invalid
    * component.  Note that this mechanism only works while this object does
    * not support multi-table queries.
    */
   private void invalidate()
   {
      assert getTableCount() == 1;
      
      Iterator<Component> iterator = components();
      if (iterator.hasNext())
      {
         AdaptiveComponent comp = (AdaptiveComponent) iterator.next();
         invalidate(comp);
      }
   }
   
   /**
    * Begin the transition of this query from dynamic to preselect mode.  The
    * remaining work of this transition is performed in the {@link #fetch}
    * and {@link #execute} methods.
    *
    * @param   breakValue
    *          The sort band break value which indicates that the current,
    *          dynamic query has crossed a sort band boundary.
    */
   private void revalidate(Object breakValue)
   {
      // Prevent thrashing between modes too often.  If we are attempting to
      // revalidate more than 3 times, we are probably in a use case which is
      // better off staying in dynamic mode.
      // TODO:  make this heuristic more intelligent, possibly by taking into
      // account the frequency of mode changes and the overall number of
      // records traversed.
      if (++revalCount > 3)
      {
         if (revalCount == 4 && LOG.isDebugEnabled())
         {
            LOG.debug("AdaptiveQuery permanently forced to dynamic mode:  " +
                      toString());
         }
         
         return;
      }
      
      this.breakValue = breakValue;
      adapter.prepareDelegateChange();
      
      if (LOG.isTraceEnabled())
      {
         String hql = getHQL();
         
         if (hql != null)
         {
            hql = hql.replaceAll("\t\n\r", "\\\\t\\\\n\\\\r");
         }
         
         Exception exc = new Exception("Details (not an error):  [" + 
                                       hql + "]");         
                                       
         LOG.trace("Adaptive query reset to preselect mode at record:  " +
                   invalidComponent.getBuffer().getDMOInterface().getName() +
                   "#" +
                   adapter.getID(0) +
                   ";  breakValue = " + breakValue,
                   exc);
      }
   }
   
   /**
    * A query component implementation which extends the
    * <code>Component</code> inner class of <code>PreselectQuery</code> to
    * store additional information needed when reverting to dynamic mode.
    */
   protected class AdaptiveComponent
   extends Component
   {
      /** HQL sort clause associated with this query component, original */
      private final String originalSort;
      
      /** DMO to which this component should join via a foreign relation. */
      private final DataModelObject inverse;
      
      /** Sort criteria associated with this query component */
      private List<SortCriterion> sortCriteria = null;
      
      /** DMO sorter associated with this component's sort clause */
      private DMOSorter dmoSorter = null;
      
      /** HQL preprocessor used for dynamic mode */
      private HQLPreprocessor dynamicHQLPreprocessor;
      
      /** Substitution arguments for this component when in dynamic mode */
      private Object[] dynamicArgs = null;
      
      /** Event registration ID used to gather index modification events */
      private Long globalEventID = null;
      
      /** ID of last index modification event retrieved from clearinghouse */
      private long lastGlobalEventID = -1L;
      
      /**
       * Constructor.
       *
       * @param   proxy
       *          Reference to record buffer updated by this component of the
       *          query.
       * @param   join
       *          Helper object for a join via a foreign key relation.  May
       *          be <code>null</code>.
       * @param   where
       *          Restriction criteria in HQL where clause format.
       * @param   sort
       *          Sort criteria in HQL order by clause format.
       * @param   indexInfo
       *          Index information string as it is returned by INDEX-INFORMATION.
       *          May be <code>null</code> if it is not an OPEN QUERY case or if
       *          you don't need debug information about selected indexes.
       * @param   lockType
       *          Requested record lock type.
       * @param   args
       *          Default (unresolved) query substitution arguments.
       * @param   iteration
       *          Type of iteration:  FIRST, LAST, NEXT, PREVIOUS, or UNIQUE.
       * @param   outer
       *          <code>true</code> if this component is joined with the
       *          previous component via an outer join;  <code>false</code>
       *          if there is no previous component, or if this component is
       *          joined with the previous component via an inner join.
       * @param   inverse
       *          DMO to which this query should join via a foreign relation.
       *          May be <code>null</code>.
       */
      AdaptiveComponent(BufferReference proxy,
                        AbstractJoin join,
                        String where,
                        String sort,
                        String indexInfo,
                        LockType lockType,
                        Object[] args,
                        int iteration,
                        boolean outer,
                        BufferReference inverse)
      {
         super(proxy, join, where, indexInfo, lockType, args, iteration, outer);
         
         this.originalSort = sort;
         this.inverse = inverse;
      }
      
      /**
       * Get the HQL preprocessor associated with this component.  This may
       * return different preprocessors, depending upon whether the query is
       * currently in preselect vs. dynamic mode.
       * 
       * @return  HQL preprocessor.
       */
      protected HQLPreprocessor getHQLPreprocessor()
      throws PersistenceException
      {
         HQLPreprocessor preselectPreproc = super.getHQLPreprocessor();
         
         return (isPreselect()? preselectPreproc : dynamicHQLPreprocessor);
      }
      
      /**
       * Create the <code>HQLPreprocessor</code> which backs this component.
       * <p>
       * This method overrides the parent's implementation to also create a
       * secondary HQL preprocessor for use in dynamic fetch mode.  This
       * secondary version does not inline any subtitution parameters.  If the
       * primary version did not inline any such parameters, then both the
       * primary and secondary preprocessors will be the same object.
       * 
       * @param   queryArgs
       *          Query substitution parameters used in the preprocessor's
       *          creation.
       * 
       * @return  Primary HQL preprocessor (i.e., the version appropriate for
       *          use with the enclosing query's preselect mode).
       * 
       * @throws  PersistenceException
       *          if there is an error creating the preprocessor.
       */
      protected HQLPreprocessor
      prepareHQLPreprocessor(Object[] queryArgs)
      throws PersistenceException
      {
         HQLPreprocessor preselectPreproc =
            super.prepareHQLPreprocessor(queryArgs);
         
         // If HQL preprocessor was inlined for preselect use, we need a
         // non-inlined version for dynamic use.
         if (preselectPreproc.wasInlined())
         {
            dynamicHQLPreprocessor =
               HQLPreprocessor.get(getOriginalWhere(),
                                   getBuffer(),
                                   queryArgs,
                                   false);
         }
         else
         {
            dynamicHQLPreprocessor = preselectPreproc;
         }
         
         return preselectPreproc;
      }
      
      /**
       * Get substitution arguments for this component, resolving them first
       * if necessary.  This may return different arguments, depending upon
       * whether the query is in preselect or dynamic mode.
       *
       * @return  Substitution arguments.
       */
      protected Object[] getArgs()
      throws PersistenceException
      {
         Object[] filteredArgs;
         
         if (isPreselect())
         {
            filteredArgs = super.getArgs();
         }
         else
         {
            if (dynamicArgs == null && getRawArgs() != null)
            {
               if (super.getHQLPreprocessor().wasInlined())
               {
                  dynamicArgs = filterArgs(dynamicHQLPreprocessor);
               }
               else
               {
                  dynamicArgs = super.getArgs();
               }
            }
            
            filteredArgs = dynamicArgs;
         }
         
         return filteredArgs;
      }
      
      /**
       * Reset both the preselect and dynamic HQL preprocessors, so that they
       * need to be recreated the next time one is requested.
       */
      protected void resetHQLPreprocessor()
      {
         super.resetHQLPreprocessor();
         dynamicHQLPreprocessor = null;
      }
      
      /**
       * Initialize instance variables which could not be set at construction,
       * because they required a scope to first be opened on the backing
       * buffer.
       * 
       * @throws  PersistenceException
       *          if there is a database or parsing error.
       */
      void initialize()
      throws PersistenceException
      {
         if (sortCriteria != null)
         {
            return;
         }
         
         RecordBuffer buffer = getBuffer();
         
         buffer.initialize();
         
         sortCriteria =
            Collections.unmodifiableList(SortCriterion.parse(originalSort,
                                                             buffer));
         
         SortIndex sortIndex = SortIndex.get(originalSort, buffer);
         setSortIndex(sortIndex);
         
         IndexHelper indexHelper = buffer.getIndexHelper();
         dmoSorter = indexHelper.getSorterForSortPhrase(buffer, originalSort);
      }
      
      /**
       * Collect and process global events which describe changes made to DMO
       * objects in other sessions.  The processing of such events may cause
       * the containing query's current results to be invalidated.
       */
      void processGlobalEvents()
      {
         // Register to receive events if necessary.
         if (globalEventID == null)
         {
            registerForGlobalEvents();
         }
         
         // Gether and process events.
         try
         {
            DirtyShareContext dirtyContext = getBuffer().getDirtyContext();
            GlobalChangeEvent[] events =
               dirtyContext.getGlobalEvents(globalEventID, lastGlobalEventID);
            
            if (events == null)
            {
               return;
            }
            
            int len = events.length;
            if (len > 0)
            {
               lastGlobalEventID = events[len - 1].getEventID();
            }
            
            for (int i = 0; i < len && isPreselect(); i++)
            {
               stateChanged(events[i]);
            }
         }
         catch (EventRegistrationException exc)
         {
            invalidate(this);
            registerForGlobalEvents();
         }
         catch (PersistenceException exc)
         {
            ErrorManager.recordOrThrowError(exc);
         }
      }
      
      /**
       * Get the list of sort criteria associated with this query component.
       *
       * @return  An unmodifiable list of sort components.
       */
      List<SortCriterion> getSortCriteria()
      {
         return sortCriteria;
      }
      
      /**
       * Get sort criteria in HQL order by clause format for this component.
       *
       * @return  HQL order by clause, before being rewritten by the
       *          <code>SortCriterion</code> class.
       */
      String getOriginalSort()
      {
         return originalSort;
      }
      
      /**
       * Get DMO (proxy) to which this component should join via a foreign
       * relation.
       *
       * @return  Inverse DMO proxy or <code>null</code> if no such join is
       *          defined for this query.
       */
      DataModelObject getInverse()
      {
         return inverse;
      }
      
      /**
       * Get the DMO sorter associated with this the index used to sort for
       * this query component.
       * 
       * @return  DMO sorter associated with this component's index.
       */
      DMOSorter getDMOSorter()
      {
         return dmoSorter;
      }
      
      /**
       * Clean up when this component is no longer in use.  This involves
       * deregistering for global events, if the component was registered for
       * the same.
       */
      void cleanup()
      {
         if (globalEventID != null)
         {
            DirtyShareContext dirtyContext = getBuffer().getDirtyContext();
            dirtyContext.deregisterForGlobalEvents(globalEventID);
         }
      }
      
      /**
       * Register to receive global DMO change events, which may cause the
       * current result set of the containing query to be invalidated.
       */
      private void registerForGlobalEvents()
      {
         DirtyShareContext dirtyContext = getBuffer().getDirtyContext();
         String entity = getBuffer().getEntityName();
         globalEventID = dirtyContext.registerForGlobalEvents(entity);
         lastGlobalEventID = globalEventID;
      }
   }
   
   /**
    * An implementation of the {@link Results} interface which uses a single
    * table or multi-table <code>DynamicQuery</code> as a delegate.  An
    * instance of this method replaces the enclosing query's preselected
    * results when a DMO state change invalidates those results.  If a single
    * table must be accessed, a modified <code>RandomAccessQuery</code> is
    * used as a delegate.  For multiple tables, a <code>CompoundQuery</code>
    * is used.
    * <p>
    * The adaptive query's original query components are interrogated for
    * their properties in order to create the dynamic query (and its
    * components in the multi-table case).
    * <p>
    * <b>WARNING:</b>  currently, the multi-table implementation is not
    * functional.  The <code>CompoundQuery</code> instance which is created
    * does not have the proper state to continue driving the query where the
    * preselect query left off.  There is a mismatch between the "flat" model
    * of the preselect result set and the "nested looping" model of the
    * <code>CompoundQuery</code> implementation which has not yet been
    * resolved.
    */
   private class DynamicResults
   implements Results
   {
      /** Underlying query which manages the navigation of results */
      private final DynamicQuery query;
      
      /** All buffers being updated by the dynamic query */
      private final RecordBuffer[] buffers = getRecordBuffers();
      
      /** ID-only flags, ordered by column */
      private final boolean[] idOnly;
      
      /** Flag indicating this results object has been closed */
      private boolean closed = false;
      
      /**
       * Constructor which accepts a backing, dynamic query and turns
       * scrolling on for that query.
       *
       * @param   query
       *          Dynamic query which will back this results object.
       * 
       * @throws  PersistenceException
       *          if there is an error beginning a new transaction.
       */
      DynamicResults(DynamicQuery query, boolean[] idOnly)
      throws PersistenceException
      {
         this.query = query;
         this.idOnly = idOnly;
         query.setScrolling();
         
         getPersistence().pushImplicitTransaction();
      }
      
      /**
       * Move cursor to the first results row.
       *
       * @return  <code>true</code> if there are any results.
       */
      public boolean first()
      {
         query.first();
         
         return postFetch(false);
      }
      
      /**
       * Move cursor to the last results row.
       *
       * @return  <code>true</code> if there are any results.
       */
      public boolean last()
      {
         query.last();
         
         return postFetch(true);
      }
      
      /**
       * Move cursor to the next results row.
       *
       * @return  <code>true</code> if there is a result under the cursor
       *          under the move.
       */
      public boolean next()
      {
         query.next();
         
         return postFetch(true);
      }
      
      /**
       * Move cursor to the previous results row.
       *
       * @return  <code>true</code> if there is a result under the cursor
       *          under the move.
       */
      public boolean previous()
      {
         query.previous();
         
         return postFetch(false);
      }
      
      /**
       * Is the cursor on the first row in the results set?
       *
       * @return  <code>true</code> if the cursor is on the first row.
       */
      public boolean isFirst()
      {
         return (getRowNumber() == 0);
      }
      
      /**
       * Is the cursor on the last row in the results set?
       *
       * @return  <code>true</code> if the cursor is on the last row.
       */
      public boolean isLast()
      {
         throw new UnsupportedOperationException();
      }
      
      /**
       * Get the array of objects at the current result row.
       *
       * @return  Object array or <code>null</code>.
       */
      public Object[] get()
      {
         int len = buffers.length;
         Object[] row = new Object[len];
         for (int i = 0; i < len; i++)
         {
            row[i] = get(i);
         }
         
         return row;
      }
      
      /**
       * Get the object at the current result row, at the specified column.
       *
       * @param   column
       *          Zero-based index column of the desired object.
       *
       * @return  Object at <code>column</code> or <code>null</code>.
       */
      public Object get(int column)
      {
         Persistable dmo = buffers[column].getCurrentRecord();
         
         return ((dmo != null && idOnly[column]) ? dmo.getId() : dmo);
      }
      
      /**
       * Get the primary key ID at the current result row, at the specified
       * column.
       *
       * @param   column
       *          Zero-based index column of the desired ID.
       *
       * @return  ID of the record or <code>null</code>.
       */
      public Serializable getID(int column)
      {
         Object datum = get(column);
         
         return ((datum != null && idOnly[column])
                 ? (Serializable) datum
                 : ((Persistable) datum).getId());
      }
      
      /**
       * Get the row number currently under the cursor.
       *
       * @return  Zero-based index of the current row, or <code>0</code> by
       *          default if the cursor's position is not currently known.
       */
      public int getRowNumber()
      {
         integer rowNum = query.currentRowImpl();
         
         return (rowNum.isUnknown() ? 0 : rowNum.intValue());
      }
      
      /**
       * Set the row number currently under the cursor.
       *
       * @param   row
       *          Zero-based index of the row to be set as the current row.
       *
       * @return  <code>true</code> if there is a row at the specified row
       *          number;  else <code>false</code>.
       */
      public boolean setRowNumber(int row)
      {
         reposition(row);
         
         return postFetch(false);
      }
      
      /**
       * Scroll the cursor ahead by the specified number of rows.  If the
       * number is negative, the cursor is moved backward.  This may put the
       * cursor off the end (either end) of the query's results.
       *
       * @param   rows
       *          Number of rows to jump ahead or back.
       *
       * @return  <code>true</code> if there is a row at the new location;
       *          else <code>false</code>.
       */
      public boolean scroll(int rows)
      {
         // Set the delegate query's fetchOnReposition strategy to match the
         // enclosing AdaptiveQuery's current strategy.
         query.setFetchOnReposition(isFetchOnReposition());
         
         if (rows < 0)
         {
            query.backward(-rows);
         }
         else
         {
            query.forward(rows);
         }
         
         return postFetch(rows > 0);
      }
      
      /**
       * Reset the cursor to its natural starting position, before the first
       * result row.
       */
      public void reset()
      {
         reposition(0);
      }
      
      /**
       * Invoked when the current Hibernate session is about to close.  Gives
       * the implementation an opportunity to perform appropriate processing in
       * response to this event.
       */
      public void sessionClosing()
      {
      }
      
      /**
       * Clean up and release any resources which this object is holding, such
       * as open result sets.
       */
      public void cleanup()
      {
         if (!closed)
         {
            closed = true;
            getPersistence().popImplicitTransaction();
         }
      }

      /**
       * Reposition the backing query to the given row.  This method sets the
       * fetch-on-reposition strategy of the backing query to that of the
       * enclosing query before attempting the reposition.
       *
       * @param   row
       *          Zero-based index of the target row.
       */
      private void reposition(int row)
      {
         // Set the delegate query's fetchOnReposition strategy to match the
         // enclosing AdaptiveQuery's current strategy.
         query.setFetchOnReposition(isFetchOnReposition());
         
         query.reposition(row + 1);    // reposition() is 1-based
      }
      
      /**
       * Optionally wwitch the enclosing back to preselect mode if possible
       * and provide a return value which indicates whether the results object
       * currently is positioned on a row.
       * 
       * @param   revalidate
       *          <code>true</code> if query should attempt to switch back to
       *          preselect mode, else <code>false</code>.  The switch should
       *          only be made if the last operation moved forward through the
       *          query's results.
       *
       * @return  <code>true</code> if positioned on a row, else
       *          <code>false</code> if off-end.
       */
      private boolean postFetch(boolean revalidate)
      {
         if (revalidate)
         {
            Object breakValue = query.getBreakValue();
            if (breakValue != null)
            {
               revalidate(breakValue);
            }
         }
         
         return !query._isOffEnd();
      }
   }
   
   /**
    * A {@link ResultsProvider} implementation which is invoked when
    * transitioning between preselect and dynamic modes (both directions).
    * Provides a callback for {@link ResultsAdapter} when a new delegate
    * {@link Results delegate worker} is needed by the adapter.
    */
   private class ResultsSource
   implements ResultsProvider
   {
      /**
       * Invoked by a {@link ResultsAdapter} when a new <code>Results</code>
       * delegate worker is needed.  Determines whether we need a preselect
       * or dynamic results delegate and installs the appropriate delegate in
       * the results adapter.
       *
       * @param   adapter
       *          <code>ResultsAdapter</code> which is the source of the
       *          request.
       */
      public void provideNewResults(ResultsAdapter adapter)
      {
         // Dynamic --> Preselect transition.
         if (breakValue != null)
         {
            // Allows execute() to determine that we now are operating in
            // preselect mode.
            invalidComponent = null;

            // Allows super.execute() to determine that we need a new
            // preselect result set.
            adapter.cleanup();
            setResults(null);

            // Retrieve preselect result set.  Adapter will be reconnected to
            // new delegate results.
            execute();

            // Store the last record(s) retrieved in dynamic mode.
            int i = 0;
            boolean recordsPresent = true;
            Serializable[] ids = new Serializable[getTableCount()];
            Iterator<Component> iter = components();
            while (iter.hasNext())
            {
               Component comp = iter.next();
               RecordBuffer buf = comp.getBuffer();
               if (buf.getCurrentRecord() != null)
               {
                  ids[i++] = buf.getCurrentRecord().getId();
               }
               else
               {
                  recordsPresent = false;
                  break;
               }
            }

            // Get the first result. It could be already visited into the
            // dynamic mode and we should skip it.
            if (recordsPresent && adapter.first())
            {
               Object[] res = adapter.get();
               for (int j = 0; j < res.length; j++)
               {
                  Serializable id;
                  if (res[j] instanceof Persistable)
                  {
                     id = ((Persistable)res[j]).getId();
                  }
                  else
                  {
                     id = (Serializable)res[j];
                  }

                  if (!ids[j].equals(id))
                  {
                     // We didn't visit this result, reset to the original
                     // position.
                     adapter.reset();
                     break;
                  }
               }
            }
         }
         // Preselect --> Dynamic transition.
         else
         {
            // Allows execute() to determine that we need a new
            // DynamicResults object.
            adapter.setResults(null);
            
            // Retrieve dynamic results.  Adapter will be reconnected to new
            // delegate results.
            execute();
         }
      }
   }
}
